import torch


class Initializer(object):
    def __call__(self, module: torch.nn.Module):
        raise NotImplementedError()

    def __str__(self):
        return "UnknownInitializer"


class Norm001Initializer(Initializer):
    def __call__(self, m : torch.nn.Module):
        if type(m) in [torch.nn.Linear, pypaddle.sparse.MaskedLinearLayer]:
            torch.nn.init.normal_(m.weight, 0, 0.1)
            m.bias.data.fill_(0.01)

    def __str__(self):
        return "Norm(0; 0.1)"



from pyklopp import __version__
from cleo import Application as BaseApplication

from .commands.init import InitCommand
from .commands.train import TrainCommand
from .commands.eval import EvalCommand


class Application(BaseApplication):
    def __init__(self):
        super(Application, self).__init__(
            'pyklopp', __version__
        )

        for command in self.get_default_commands():
            self.add(command)

    def get_default_commands(self):  # type: () -> list
        commands = [
            InitCommand(),
            TrainCommand(),
            EvalCommand()
        ]

        return commands




import os
import sys
import torch


def count_parameters(model: torch.nn.Module):
    return sum(p.numel() for p in model.parameters() if p.requires_grad)


def subpackage_import(name: str):
    components = name.split('.')

    mod = __import__(components[0])
    for comp in components[1:]:
        mod = getattr(mod, comp)

    return mod


def add_local_path(fn_info=None):
    # Add current absolute path to system path to load local modules
    # If initialized a module previously from a local module, then it must be available in path later again
    add_path = os.path.abspath('.')
    if add_path not in sys.path:
        sys.path.append(add_path)

        if fn_info is not None:
            fn_info('Added "%s" to path.' % add_path)


def load_modules(module_args : list):
    loaded_modules = []

    if module_args is None:
        module_args = []

    for module_option in module_args:
        module_option = str(module_option)
        possible_module_file_name = module_option + '.py' if not module_option.endswith('.py') else module_option
        if os.path.exists(possible_module_file_name):
            module_file_name = possible_module_file_name
            module_name = module_file_name.replace('.py', '')

            try:
                loaded_modules.append(__import__('.' + module_name, fromlist=['']))
            except ModuleNotFoundError:
                raise ModuleNotFoundError('Could not import "%s"' % module_name)

    return loaded_modules


def load_dataset_from_argument(dataset_arg : str, assembled_config : dict) -> torch.utils.data.Dataset:
    """

    :param dataset_arg:
    :param assembled_config:
    :return:
    """
    fn_get_dataset = None  # optional function to load the dataset based on the allocated configuration
    class_dataset = None  # optional class which will be instanatiated with the configuration sub key 'dataset_config'

    add_local_path()

    # For bash-completion, also allow the module name to end with .py and then simply remove it
    dataset_module_file_name = None
    dataset_possible_module_file_name = dataset_arg + '.py' if not dataset_arg.endswith('.py') else dataset_arg
    if os.path.exists(dataset_possible_module_file_name):
        # We've got an argument for which a local file exists, e.g. 'my_module' or 'my_module.py'
        dataset_module_file_name = dataset_possible_module_file_name
        dataset_module_name = dataset_module_file_name.replace('.py', '')

        try:
            module = __import__('.' + dataset_module_name, fromlist=[''])
        except ModuleNotFoundError:
            raise ModuleNotFoundError('Could not import "%s"' % dataset_module_name)

        try:
            fn_get_dataset = module.get_dataset
        except AttributeError:
            raise ValueError('Could not find dataset() function in your module "%s". You probably need to define get_dataset(**kwargs)' % dataset_module_name)

    # Assume argument to be s.th. like 'torchvision.datasets.cifar.CIFAR10'
    if dataset_module_file_name is None:
        if not '.' in dataset_arg:
            raise ValueError('Dataset must be an attribute of a module. Expecting at least one dot.')
        try:
            class_dataset = subpackage_import(dataset_arg)
        except ModuleNotFoundError:
            raise ValueError('Could not import %s' % dataset_arg)

    if fn_get_dataset is None and class_dataset is None:
        raise ValueError('Neither a dataset class nor a get_dataset(**kwargs) is defined.')

    if fn_get_dataset:
        # The preferred way to load a dataset: via a getter function my_module.get_dataset(**config) -> torch.utils.data.Dataset
        dataset = fn_get_dataset(**assembled_config)

        assembled_config['dataset_getter'] = fn_get_dataset.__name__
    else:
        dataset_config = {key.replace('dataset_', ''): assembled_config[key] for key in assembled_config if key.startswith('dataset_')}

        # In case of class instanatiation, try to load the custom transformation function
        if 'get_dataset_transformation' in assembled_config:
            get_dataset_transformation = assembled_config['get_dataset_transformation']
            try:
                fn_get_custom_transformation = subpackage_import(get_dataset_transformation)
            except ModuleNotFoundError:
                raise ValueError('Could not import transformation %s' % get_dataset_transformation)
            except AttributeError as e:
                raise ValueError('Could not load transformation due to attribute error for %s: %s' % (get_dataset_transformation, e))

            dataset_config['transform'] = fn_get_custom_transformation()

        dataset = class_dataset(**dataset_config)

        assembled_config['dataset_class'] = class_dataset.__name__

    return dataset



import json
import os
import random
import socket
import sys
import time
import uuid
import torch
import numpy as np
import ignite

from ignite.contrib.handlers import ProgressBar
from ignite.engine import create_supervised_evaluator
from cleo import Command

from pyklopp import __version__, subpackage_import
from pyklopp.util import load_modules, load_dataset_from_argument, add_local_path


class EvalCommand(Command):
    """
    Evaluates a pre-trained model on a given test data set.

    eval
        {model : Path to the pytorch model file}
        {testset : Function to retrieve the test set based on the assembled configuration}
        {--m|modules=* : Optional module file to load.}
        {--c|config=* : Configuration JSON string or file path.}
        {--s|save= : Path (including optional name) to save the configuration to, e.g. sub/path/config.json}
    """

    def handle(self):
        """
            Early path checking. This helps detecting errors early before loading a model or a dataset.
        """
        # Early check for model path
        model_path = str(self.argument('model'))
        if not os.path.exists(model_path):
            raise ValueError('No model file found on "%s"' % model_path)

        # Early check for save path
        existing_config = None
        save_path_base = None
        config_file_name = None
        if self.option('save'):
            save_path = str(self.option('save'))
            if os.path.exists(save_path):
                with open(save_path, 'r') as handle:
                    existing_config = json.load(handle)

            config_file_name = os.path.basename(save_path)
            save_path_base = os.path.dirname(save_path)
            if not os.path.isdir(save_path_base):
                os.makedirs(save_path_base)

        # Load optional modules
        add_local_path(self.info)
        modules_option = self.option('modules')
        loaded_modules = load_modules(modules_option)

        """
            Assemble configuration
        """
        dataset_argument = str(self.argument('testset'))
        config = {
            'global_unique_id': str(uuid.uuid4()),
            'pyklopp_version': __version__,
            'loaded_modules': loaded_modules,
            'python_seed_initial': None,
            'python_seed_random_lower_bound': 0,
            'python_seed_random_upper_bound': 10000,
            'python_cwd': os.getcwd(),
            'hostname': socket.gethostname(),
            'time_config_start': time.time(),
            'model_path': model_path,
            'save_path_base': save_path_base,
            'config_persistence_name': 'config.json',
            'config_key': 'evaluation',
            'gpus_exclude': [],
            'gpu_choice': None,  # if None, then random uniform of all available is chosen
            'batch_size': 100,
            'argument_dataset': dataset_argument,
            'get_dataset_transformation': 'pyklopp.defaults.get_transform',
            'get_loss': 'pyklopp.defaults.get_loss'
        }

        # Load user-defined configuration
        if self.option('config'):
            for config_option in self.option('config'):
                config_option = str(config_option)
                if os.path.exists(config_option):
                    self.info('Loading configuration from "%s"' % config_option)
                    with open(config_option, 'r') as handle:
                        user_config = json.load(handle)
                else:
                    try:
                        user_config = json.loads(config_option)
                    except TypeError:
                        raise ValueError('Invalid JSON as config passed.')
                assert type(user_config) is dict, 'User config must be a dictionary.'

                config.update(user_config)

        # If desired set an initial (global) random seed
        if config['python_seed_initial'] is not None:
            random.seed(int(config['python_seed_initial']))

        # Generate a (local) initial seed, which might depend on the initial global seed
        # This enables reproducibility
        a = config['python_seed_random_lower_bound']
        b = config['python_seed_random_upper_bound']
        python_seed_local = random.randint(a, b)
        config['python_seed_local'] = python_seed_local if 'python_seed_local' not in config else config['python_seed_local']
        random.seed(config['python_seed_local'])
        np.random.seed(config['python_seed_local'])
        torch.manual_seed(config['python_seed_local'])
        torch.cuda.manual_seed(config['python_seed_local'])

        """
        Load configured dataset for evaluation.
        """
        self.info('Loading dataset.')
        dataset = load_dataset_from_argument(dataset_argument, config)

        config['dataset'] = str(dataset.__class__.__name__)
        n_samples = len(dataset)
        train_sampler = torch.utils.data.SubsetRandomSampler(np.arange(n_samples, dtype=np.int64))
        config['time_dataset_loading_start'] = time.time()
        dataset_loader = torch.utils.data.DataLoader(
            dataset,
            batch_size=config['batch_size'],
            sampler=train_sampler,
            num_workers=2
        )
        dataiter = iter(dataset_loader)
        initial_features, initial_labels = dataiter.next()
        config['time_dataset_loading_end'] = time.time()

        # Determine device to use (e.g. cpu, gpu:0, gpu:1, ..)
        if torch.cuda.is_available():
            cuda_no = config['gpu_choice']
            if cuda_no is None:
                cuda_no = np.random.choice(np.setdiff1d(np.arange(torch.cuda.device_count()), config['gpus_exclude']))
            elif cuda_no in config['gpus_exclude']:
                raise ValueError('Your configured GPU device number is in the exclusion list of your configuration.')
            device = torch.device('cuda:%s' % cuda_no)
        else:
            device = torch.device('cpu')
        config['device'] = str(device)

        # Load the model
        try:
            model = torch.load(model_path, map_location=device)
        except ModuleNotFoundError as e:
            raise ValueError('Could not find module when loading model: %s' % e)
        model.to(device)
        config['model_pythonic_type'] = str(type(model))

        fn_get_loss = subpackage_import(config['get_loss'])
        fn_loss = fn_get_loss(**config)
        config['loss'] = str(fn_loss.__class__.__name__)

        self.info('Configuration:')
        self.info(json.dumps(config, indent=2, sort_keys=True))

        """
            Evaluation
        """
        metric_accuracy = ignite.metrics.Accuracy()
        metric_precision = ignite.metrics.Precision(average=False)
        metric_recall = ignite.metrics.Recall(average=False)
        metric_f1 = (metric_precision * metric_recall * 2 / (metric_precision + metric_recall)).mean()
        evaluation_metrics = {
            'accuracy': metric_accuracy,
            'precision': ignite.metrics.Precision(average=True),
            'recall': ignite.metrics.Recall(average=True),
            'f1': metric_f1,
            'loss': ignite.metrics.Loss(fn_loss)
        }
        evaluator = create_supervised_evaluator(
            model,
            metrics=evaluation_metrics,
            device=device
        )

        config['time_model_evaluation_start'] = time.time()
        evaluation_state = evaluator.run(dataset_loader)
        config['time_model_evaluation_end'] = time.time()

        for metric_name in evaluation_metrics:
            config['evaluation_%s' % metric_name] = np.float(evaluation_state.metrics[metric_name])

        self.info('Final configuration:')
        self.info(json.dumps(config, indent=2, sort_keys=True))
        self.info('Done.')


import random
import socket
import sys
import os
import json
import time
import uuid
import torch.nn
import numpy as np

from cleo import Command
from pyklopp import __version__
from pyklopp import subpackage_import
from pyklopp.util import count_parameters


class InitCommand(Command):
    """
    Initializes a model from a given module

    init
        {model : Name of the module with initialization method for the model.}
        {--m|modules=* : Optional modules to load.}
        {--c|config=* : JSON config or path to JSON config file.}
        {--s|save= : Path to save the model/config to}
    """

    def handle(self):
        # Early check for save path
        save_path_base = None
        model_file_name = None
        if self.option('save'):
            save_path = str(self.option('save'))
            model_file_name = os.path.basename(save_path)
            save_path_base = os.path.dirname(save_path)
            if len(save_path_base) < 1:
                raise ValueError('You did not specify a path with "%s"' % save_path)
            if os.path.exists(os.path.join(save_path_base, model_file_name)):
                raise ValueError('Path "%s" already exists' % save_path_base)

            if not os.path.exists(save_path_base):
                os.makedirs(save_path_base)

        # Add current absolute path to system path
        # This is required for local modules to load.
        add_path = os.path.abspath('.')
        sys.path.append(add_path)
        self.info('Added %s to path' % add_path)

        """
        Optional (local) module to load.
        There several functionalities can be bundled at one place.
        """
        modules_option = self.option('modules')
        loaded_modules = []
        if modules_option:
            for module_option in modules_option:
                possible_module_file_name = module_option + '.py' if not module_option.endswith('.py') else module_option
                if os.path.exists(possible_module_file_name):
                    module_file_name = possible_module_file_name
                    module_name = module_file_name.replace('.py', '')

                    try:
                        loaded_modules.append(__import__('.' + module_name, fromlist=['']))
                    except ModuleNotFoundError:
                        raise ModuleNotFoundError('Could not import "%s"' % module_name)
        self.info('Loaded modules: %s' % loaded_modules)


        """ ---------------------------
        Static initial configuration.
        This can be overwritten by user-defined configurations.
        Dynamic computed configurations can be configured by some parameters and are computed after the user-defined
        values have been set.
        """
        config = {
            'global_unique_id': str(uuid.uuid4()),
            'pyklopp_version': __version__,
            'loaded_modules': loaded_modules,
            'gpus_exclude': [],
            'python_seed_initial': None,
            'python_seed_random_lower_bound': 0,
            'python_seed_random_upper_bound': 10000,
            'python_cwd': os.getcwd(),
            'hostname': socket.gethostname(),
            'time_config_start': time.time(),
            'save_path_base': save_path_base,
            'model_persistence_name': model_file_name,  # If later set to None/empty, model will not be persisted
        }

        # Load user-defined configuration
        if self.option('config'):
            for config_option in self.option('config'):
                config_option = str(config_option)
                if os.path.exists(config_option):
                    self.info('Loading configuration from "%s"' % config_option)
                    with open(config_option, 'r') as handle:
                        user_config = json.load(handle)
                else:
                    try:
                        user_config = json.loads(config_option)
                    except TypeError:
                        raise ValueError('Invalid JSON as config passed.')
                assert type(user_config) is dict, 'User config must be a dictionary.'

                config.update(user_config)

        """ ---------------------------
        Dynamic configuration computations.
        E.g. random numbers or seeds etc.
        """

        # If desired set an initial (global) random seed
        if config['python_seed_initial'] is not None:
            random.seed(int(config['python_seed_initial']))

        # Generate a (local) initial seed, which might depend on the initial global seed
        # This enables reproducibility
        a = config['python_seed_random_lower_bound']
        b = config['python_seed_random_upper_bound']
        python_seed_local = random.randint(a, b)
        config['python_seed_local'] = python_seed_local if 'python_seed_local' not in config else config['python_seed_local']
        random.seed(config['python_seed_local'])
        np.random.seed(config['python_seed_local'])
        torch.manual_seed(config['python_seed_local'])
        torch.cuda.manual_seed(config['python_seed_local'])

        config['time_config_end'] = time.time()


        # For ease of usage, get_model can be either 'my_module.py' containing a get_model(**config) function
        # or it can be a function such as my_module.get_model with signature (**config) where my_module can be
        # a loadable module in the local path or an already loaded module via --modules
        get_model = self.argument('model')
        get_model_components = get_model.split('.')
        fn_get_model = None  # This will be the model loading function with signature fn(**config)

        if get_model.endswith('.py') or os.path.exists(get_model_components[0] + '.py'):
            module_name = get_model_components[0]
            module_file_name = module_name + '.py'
            if not os.path.exists(module_file_name):
                raise ValueError('No such local file %s' % module_file_name)

            try:
                module = __import__(module_name, fromlist=[''])
            except ModuleNotFoundError:
                raise ModuleNotFoundError('Could not import module "' + module_name + '"')

            if get_model.endswith('.py') or len(get_model_components) < 2:
                # get_model was 'my_module.py' or 'my_module'
                try:
                    fn_get_model = module.get_model
                except AttributeError:
                    raise ValueError(
                        'Could not find get_model() function in your module "%s". You probably need to define get_model(**config)' % module_name)
            else:
                # get_model was s.th. like 'my_module.load_model'
                mod = module
                for load_name in get_model_components[1:]:
                    mod = getattr(mod, load_name)
                fn_get_model = mod
        else:
            try:
                fn_get_model = subpackage_import(get_model)
            except ModuleNotFoundError as e:
                raise ValueError('Could not find module <%s> via subpackage import.'.format(get_model), e)

        config['argument_model'] = get_model
        config['get_model'] = str(fn_get_model.__name__)


        self.info('Configuration:')
        self.info(json.dumps(config, indent=2, sort_keys=True))

        # Instantiate model
        config['time_model_init_start'] = time.time()
        model = fn_get_model(**config)
        config['time_model_init_end'] = time.time()
        assert isinstance(model, torch.nn.Module)

        config['model_pythonic_type'] = str(type(model))
        config['model_trainable_parameters'] = count_parameters(model)

        if save_path_base is not None:
            if config['model_persistence_name'] is not None and len(config['model_persistence_name']) > 0:
                model_file_name = config['model_persistence_name']
                model_file_path = os.path.join(save_path_base, model_file_name)
                self.info('Saving to "%s"' % model_file_path)
                config['time_model_save_start'] = time.time()
                torch.save(model, model_file_path)
                config['time_model_save_end'] = time.time()

            config_file_name = 'config.json'
            config_file_path = os.path.join(save_path_base, config_file_name)
            self.info('Writing configuration to "%s"' % config_file_path)
            with open(config_file_path, 'w') as handle:
                json.dump({'init': config}, handle, indent=2, sort_keys=True)

        self.info('Final configuration:')
        self.info(json.dumps(config, indent=2, sort_keys=True))
        self.info('Done.')

